#!/usr/bin/perl -I/usr/sausalito/perl
# $Id: check_backups.pl
#
# This handler / constructor checks for the presence of local and remote backup
# files. It then updates CODB with that information and if need be creates new
# CODB 'Backup' objects..
#
# Additionally it walks through the existing 'Backup' CODB objects and removes
# all CODB 'Backup' objects for which we no longer have local or remote payloads.
#
# Lastly it makes sure that we don't have more remote tarballs than we have set
# 'maxfiles' to. If it discoveres more remote tarballs than we want, it deletes
# the oldes ones and removes the respective 'Backup' object.

$whatami = "handler";

use CCE;
my $cce = new CCE;
use POSIX qw(isalpha);
use Socket;
use Sys::Hostname;
use File::Spec;
use List::MoreUtils qw(uniq);
use Net::SFTP::Foreign;
use File::Basename;
use POSIX qw(strftime);
use DateTime;

# Helper to redirect SSH errors to /dev/null:
open $ssherr, '>', '/dev/null'; # or die "Unable to open /dev/null";

if ($whatami eq "handler") {
    $cce->connectfd();
}
else {
    $cce->connectuds();
}

# If this is a secondary in a cluster, we can save us the bother:
if (!-d "/vz/private") {
    $cce->bye('SUCCESS');
    exit 0;
}

# Get settings out of CODB to determine the target of the remote backup (if any):
@oids = $cce->find('VserverMasterNode');
($ok, $solbackupValues) = $cce->get($oids[0]);
if ($ok) {
    # Define values:
    $vzdump_enabled             = $solbackupValues->{'vzdump_enabled'};
    $rsync_enabled              = $solbackupValues->{'rsync_enabled'};
    $MN_RemoteBackupEnabled     = $solbackupValues->{'MN_RemoteBackupEnabled'};
    $MN_LocalBackupEnabled      = $solbackupValues->{'MN_LocalBackupEnabled'};
    $MN_BackupLocalLocation     = $solbackupValues->{'MN_BackupLocalLocation'};
    $targethost                 = $solbackupValues->{'MN_BackupRemoteTargetHost'};
    $targetip                   = $solbackupValues->{'MN_BackupRemoteTargetIP'};
    $path                       = $solbackupValues->{'MN_BackupRemoteLocation'};
    $user                       = $solbackupValues->{'MN_BackupRemoteTargetUser'};
    $pass                       = $solbackupValues->{'MN_BackupRemoteTargetUserPass'};
    $port                       = $solbackupValues->{'MN_BackupRemoteTargetPort'};
    $limit                      = $solbackupValues->{'RsyncSpeedLimit'};
    $my_maxfiles                = $solbackupValues->{'vzdump_maxfiles'};

    # Assemble paths:
    $local_location_vzdump = $solbackupValues->{'MN_BackupLocalLocation'} . "/vzbackup";
    $local_location_vzdump =~ s|//|/|g;
    $local_location_vzdump =~ s|/$||g;

    $remote_location_vzdump = $solbackupValues->{'MN_BackupRemoteLocation'};
    $remote_location_vzdump =~ s|//|/|g;
    $remote_location_vzdump =~ s|/$||g;

    $local_location_rsync = $solbackupValues->{'MN_BackupLocalLocation'} . "/rsync";
    $local_location_rsync =~ s|//|/|g;
    $local_location_rsync =~ s|/$||g;

    $remote_location_rsync = $solbackupValues->{'MN_BackupRemoteLocation'};
    $remote_location_rsync =~ s|//|/|g; 
    $remote_location_rsync =~ s|/$||g; 
}

# Check if there is anything to do. Are there any local or remote backups at these
# specified locations?

# Local RSYNCs:
if (-d $local_location_rsync) {
    # Local RSYNC backup directory found.
    $exec = `ls -d1 $local_location_rsync|wc -l`;
    if ($exec != "0") {
        # Local RSYNC directories found
        opendir(DIR, $local_location_rsync) or die $!;
        while (my $file = readdir(DIR)) {
            # Use a regular expression to ignore files beginning with a period
            next if ($file =~ m/^\./);
            next if ($file =~ m/^vzdump-(.*)\.tar$/);

            # Handle OpenVZ VPS's that use Simfs:
            if (-f "$local_location_rsync/$file/etc/vzdump/vps.conf") {
                # Found locally stored VPS backup config at $local_location_rsync/$file/etc/vzdump/vps.conf

                $virt_type = 'openvz';
                $vpsid = $file;
                $modify = `stat $local_location_rsync/$file/etc/vzdump/vps.conf | grep ^Modify:`;
                @mod_helper = split('\s+', $modify);
                $date = $mod_helper[1];
                @timeh = split('\.', $mod_helper[2]);
                $time = $timeh[0];
                $datetime = $date . " " . $time;

                $hostname = `cat $local_location_rsync/$file/etc/vzdump/vps.conf |grep ^HOSTNAME|cut -d = -f2 | cut -d \\\" -f2`;
                chomp($hostname);
                $ip = `cat $local_location_rsync/$file/etc/vzdump/vps.conf |grep ^IP_ADDRESS|awk -F = '{ print \$2 }' | awk -F \\\" '{ print \$2 }' | awk -F \\  '{ print \$1 }'`;
                chomp($ip);
                # Check if we already have a CODB object for this RSYNC backup:
                $path_helper = $local_location_rsync . "/" . $file;
                @my_local_backups = $cce->findx('Backup', {'vpsid' => $vpsid, 'location' => 'local', 'storage' => $path_helper, 'origin' => 'rsync', 'virttype' => $virt_type});
                if ($my_local_backups[0] eq "") {
                    # VPS $vpsid needs a CODB object:
                    if (-f "/tmp/vz_vserver_backup.lock") {
                        # No CODB object for this RSYNC backup present, but RSYNC is currently running. Laying off creation of CODB objects.
                    }
                    else {
                        # VZbackup is not running. Creating a matching CODB object for this rsync backup:
                        ($ok) = $cce->create('Backup', {
                            'ip' => $ip,
                            'hostname' => $hostname,
                            'origin' => 'rsync',
                            'vpsid' => $vpsid,
                            'virttype' => $virt_type,
                            'location' => 'local',
                            'LBtimestamp' => $datetime,
                            'layout' => 'simfs',
                            'storage' => $path_helper,
                            'timestamp' => time()
                        });
                    }
                }
            }
            # Handle OpenVZ VPS's that use Ploop:
            if (-f "$local_location_rsync/$file/vps.conf") {
                # Found locally stored VPS backup config at $local_location_rsync/$file/vps.conf

                # Make sure we have a tarball:
                $check = `ls -k1 $local_location_rsync/$file/vzpbackup*.tar|wc -l`;
                chomp($check);
                if ($check gt "0") {

                    $virt_type = 'openvz';
                    $vpsid = $file;
                    $modify = `stat $local_location_rsync/$file/vps.conf | grep ^Modify:`;
                    @mod_helper = split('\s+', $modify);
                    $date = $mod_helper[1];
                    @timeh = split('\.', $mod_helper[2]);
                    $time = $timeh[0];
                    $datetime = $date . " " . $time;

                    $hostname = `cat $local_location_rsync/$file/vps.conf |grep ^HOSTNAME|cut -d = -f2 | cut -d \\\" -f2`;
                    chomp($hostname);
                    $ip = `cat $local_location_rsync/$file/vps.conf |grep ^IP_ADDRESS|awk -F = '{ print \$2 }' | awk -F \\\" '{ print \$2 }' | awk -F \\  '{ print \$1 }'`;
                    chomp($ip);

                    # Check if we already have a CODB object for this RSYNC backup:
                    $path_helper = $local_location_rsync . "/" . $file;
                    @my_local_backups = $cce->findx('Backup', {'vpsid' => $vpsid, 'location' => 'local', 'storage' => $path_helper, 'origin' => 'rsync', 'virttype' => $virt_type});
                    if ($my_local_backups[0] eq "") {
                        # VPS $vpsid needs a CODB object:
                        if (-f "/tmp/vz_vserver_backup.lock") {
                            # No CODB object for this RSYNC backup present, but RSYNC is currently running. Laying off creation of CODB objects.
                        }
                        else {
                            # VZbackup is not running. Creating a matching CODB object for this rsync backup:
                            ($ok) = $cce->create('Backup', {
                                'ip' => $ip,
                                'hostname' => $hostname,
                                'origin' => 'rsync',
                                'vpsid' => $vpsid,
                                'virttype' => $virt_type,
                                'location' => 'local',
                                'LBtimestamp' => $datetime,
                                'layout' => 'ploop',
                                'storage' => $path_helper,
                                'timestamp' => time()
                            });
                        }
                    }
                }
            }
        }
        closedir(DIR);
    }
}

# Remote RSYNCs:
if ($user && $port && ($targethost||$targetip) && $remote_location_rsync) {
    if ($targethost) {
        $target = $user . '@' . $targethost;
    }
    else {
        $target = $user . '@' . $targetip;
    }

    $scheck = `ssh -p $port $target "ls -k1 $remote_location_rsync|wc -l"`;
    chomp($scheck);
    if ($scheck gt "0") {
        $dcheck = `ssh -p $port $target "ls -k1 $remote_location_rsync"`;
        chomp($dcheck);
        @remoteDirs = split "\n", $dcheck;
        for $rdir (@remoteDirs) {
            #
            ### Check for Simfs remote backup
            #
            $simcheck = `ssh -p $port $target "ls -k1 $remote_location_rsync/$rdir/etc/vzdump/vps.conf|wc -l"`;
            chomp($simcheck);
            if ($simcheck eq "1") {
                $vpsid = $rdir;
                $modify = `ssh -p $port $target "stat $remote_location_rsync/$rdir/etc/vzdump/vps.conf | grep ^M"`;
                @mod_helper = split('\s+', $modify);
                $date = $mod_helper[1];
                @timeh = split('\.', $mod_helper[2]);
                $time = $timeh[0];
                $datetime = $date . " " . $time;

                $XXX = '\"';
                $YYY = '\\ ';
                $hostname = `ssh -p $port $target 'cat $remote_location_rsync/$rdir/etc/vzdump/vps.conf |grep ^HOSTNAME|cut -d = -f2 | cut -d $XXX -f2'`;
                chomp($hostname);
                $ip = `ssh -p $port $target 'cat $remote_location_rsync/$rdir/etc/vzdump/vps.conf |grep ^IP_ADDRESS|awk -F = "{ print \$2 }" | cut -d $XXX -f2| awk -F $YYY "{ print \$1 }"'`;
                chomp($ip);

                $path_helper = $remote_location_rsync . "/" . $rdir;
                @my_remote_backups = $cce->findx('Backup', {'vpsid' => $vpsid, 'location' => 'remote', 'storage' => $path_helper, 'origin' => 'rsync', 'virttype' => 'openvz'});
                if ($my_remote_backups[0] eq "") {
                    # VPS $vpsid needs a CODB object:
                    if (-f "/tmp/vz_vserver_backup.lock") {
                        # No CODB object for this RSYNC backup present, but RSYNC is currently running. Laying off creation of CODB objects.
                    }
                    else {
                        # VZbackup is not running. Creating a matching CODB object for this rsync backup:
                        ($ok) = $cce->create('Backup', {
                            'ip' => $ip,
                            'hostname' => $hostname,
                            'origin' => 'rsync',
                            'vpsid' => $vpsid,
                            'virttype' => 'openvz',
                            'location' => 'remote',
                            'LBtimestamp' => $datetime,
                            'layout' => 'simfs',
                            'storage' => $path_helper,
                            'timestamp' => time()
                        });
                    }
                }
            }
            #
            ### Check for Ploop remote backup
            #
            $ploopcheck = `ssh -p $port $target "ls -k1 $remote_location_rsync/$rdir/vps.conf|wc -l"`;
            chomp($ploopcheck);
            if ($ploopcheck eq "1") {
                $tarball = `ssh -p $port $target "ls -k1 $remote_location_rsync/$rdir/vzpbackup*.tar|head -1"`;
                chomp($tarball);
                if ($tarball ne "") {
                    # Tarball found, too.
                    $vpsid = $rdir;
                    $modify = `ssh -p $port $target "stat $remote_location_rsync/$rdir/vps.conf | grep ^M"`;
                    @mod_helper = split('\s+', $modify);
                    $date = $mod_helper[1];
                    @timeh = split('\.', $mod_helper[2]);
                    $time = $timeh[0];
                    $datetime = $date . " " . $time;

                    $hostname = `ssh -p $port $target 'cat $remote_location_rsync/$rdir/vps.conf |grep ^HOSTNAME|cut -d = -f2 | cut -d $XXX -f2'`;
                    chomp($hostname);
                    $ip = `ssh -p $port $target 'cat $remote_location_rsync/$rdir/vps.conf |grep ^IP_ADDRESS|awk -F = "{ print \$2 }" | cut -d $XXX -f2| awk -F $YYY "{ print \$1 }"'`;
                    chomp($ip);

                    $path_helper = $remote_location_rsync . "/" . $rdir;
                    @my_remote_backups = $cce->findx('Backup', {'vpsid' => $vpsid, 'location' => 'remote', 'storage' => $path_helper, 'origin' => 'rsync', 'virttype' => 'openvz'});
                    if ($my_remote_backups[0] eq "") {
                        # VPS $vpsid needs a CODB object:
                        if (-f "/tmp/vz_vserver_backup.lock") {
                            # No CODB object for this RSYNC backup present, but RSYNC is currently running. Laying off creation of CODB objects.
                        }
                        else {
                            # VZbackup is not running. Creating a matching CODB object for this rsync backup:
                            ($ok) = $cce->create('Backup', {
                                'ip' => $ip,
                                'hostname' => $hostname,
                                'origin' => 'rsync',
                                'vpsid' => $vpsid,
                                'virttype' => 'openvz',
                                'location' => 'remote',
                                'LBtimestamp' => $datetime,
                                'layout' => 'ploop',
                                'storage' => $path_helper,
                                'timestamp' => time()
                            });
                        }
                    }
                }
            }
        }

    }
    #$cce->bye('SUCCESS');
    #exit 0;
}

# Now work in reverse: Walk through the 'Backup' objects and check if their payload is still present. If not, remove the CODB objects:
$proc_helper = `ps axf|grep /usr/sbin/vzdump |grep -v grep|wc -l`;
if ($proc_helper > '0') {
    # No CODB object for this tarball present, but VZdump is currently running. Laying off deletion of CODB objects.
}
else {
    # Backup currently not running. 

    # Verify local backups:
    @my_local_verify = $cce->findx('Backup', {'location' => 'local'});
    for $entry (@my_local_verify) {
        ($ok, $old_entry) = $cce->get($entry);
        # Handle vzdump tarballs:
        if ($old_entry->{'origin'} eq 'vzdump') {
            ($ok) = $cce->destroy($entry);
        }
        # Handle rsync directories:
        if ($old_entry->{'origin'} eq 'rsync') {
            $is_dir = $old_entry->{'storage'};
            if (!-d "$is_dir") {
                ($ok) = $cce->destroy($entry);
            }
        }
    }

    # Verify remote backups:
    if ($user && $port && ($targethost||$targetip) && $remote_location_rsync) {
        if ($targethost) {
            $target = $user . '@' . $targethost;
        }
        else {   
            $target = $user . '@' . $targetip;
        }

        # Check remote OpenVZ RSYNCs:
        undef @my_local_verify;
        @my_local_verify = $cce->findx('Backup', {'location' => 'remote', 'origin' => 'rsync', 'virttype' => 'openvz'});
        for $entry (@my_local_verify) {
            ($ok, $old_entry) = $cce->get($entry);

            #
            ### Check for Simfs/Ploop remote backups:
            #
            $simcheck = `ssh -p $port $target "ls -k1 $old_entry->{'storage'}/etc/vzdump/vps.conf|wc -l"`;
            chomp($simcheck);
            $ploopcheck = `ssh -p $port $target "ls -k1 $old_entry->{'storage'}/vps.conf|wc -l"`;
            chomp($ploopcheck);

            if (($simcheck ne "1") && ($ploopcheck ne "1")) {
                ($ok) = $cce->destroy($entry);
            }
        }
    }
}

$cce->bye('SUCCESS');
exit 0;

# 
# Copyright (c) 2016 Michael Stauber, SOLARSPEED.NET
# All Rights Reserved.
# 
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in 
#    the documentation and/or other materials provided with the 
#    distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its 
#    contributors may be used to endorse or promote products derived 
#    from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
# 
# You acknowledge that this software is not designed or intended for 
# use in the design, construction, operation or maintenance of any 
# nuclear facility.
# 